import { __spreadArray, __read } from 'tslib';
import sync, { cancelSync, flushSync } from 'framesync';
import { SubscriptionManager } from '../../utils/subscription-manager.js';
import { copyBoxInto } from '../geometry/copy.js';
import { applyBoxDelta, applyTreeDeltas } from '../geometry/delta-apply.js';
import { calcBoxDelta } from '../geometry/delta-calc.js';
import { createBox, createDelta } from '../geometry/models.js';
import { translateAxis } from '../geometry/operations.js';
import { isDeltaZero } from '../geometry/utils.js';
import { scaleCorrectors } from '../styles/scale-correction.js';
import { buildProjectionTransform } from '../styles/transform.js';
import { eachAxis } from '../utils/each-axis.js';

function createProjectionNode(_a) {
    _a.attachResizeListener; var defaultParent = _a.defaultParent, measureScroll = _a.measureScroll, measureViewportBox = _a.measureViewportBox, resetTransform = _a.resetTransform;
    return /** @class */ (function () {
        function ProjectionNode(parent) {
            var _this = this;
            this.children = new Set();
            this.options = {};
            this.treeScale = { x: 1, y: 1 }; // TODO Lazy-initialise
            this.updateProjection = function () {
                updateProjectionTree(_this);
            };
            this.parent = parent
                ? parent
                : defaultParent === null || defaultParent === void 0 ? void 0 : defaultParent();
            if (this.parent)
                this.parent.children.add(this);
            this.root = this.parent ? this.parent.root || this.parent : this;
            this.path = this.parent ? __spreadArray(__spreadArray([], __read(this.parent.path)), [this.parent]) : [];
        }
        ProjectionNode.prototype.destructor = function () {
            var _a;
            (_a = this.parent) === null || _a === void 0 ? void 0 : _a.children.delete(this);
            cancelSync.preRender(this.updateProjection);
        };
        /**
         * Lifecycles
         */
        ProjectionNode.prototype.mount = function (instance) {
            this.instance = instance;
        };
        ProjectionNode.prototype.willUpdate = function (shouldNotifyListeners) {
            var _a;
            if (shouldNotifyListeners === void 0) { shouldNotifyListeners = true; }
            if (this.isLayoutDirty)
                return;
            this.isLayoutDirty = true;
            this.path.forEach(function (node) {
                node.shouldResetTransform = true;
                /**
                 * TODO: Check we haven't updated the scroll
                 * since the last didUpdate
                 */
                node.updateScroll();
            });
            this.updateSnapshot();
            shouldNotifyListeners && ((_a = this.layoutWillUpdateListeners) === null || _a === void 0 ? void 0 : _a.notify());
        };
        // Note: Currently only running on root node
        ProjectionNode.prototype.didUpdate = function () {
            /**
             * Write
             */
            resetTreeTransform(this);
            /**
             * Read ==================
             */
            // Update layout measurements of updated children
            updateTreeLayout(this);
            /**
             * Write
             */
            // Notify listeners that the layout is updated
            notifyLayoutUpdate(this);
            // Flush any scheduled updates
            flushSync.update();
            flushSync.preRender();
            flushSync.render();
        };
        ProjectionNode.prototype.scheduleUpdateProjection = function () {
            sync.preRender(this.updateProjection, false, true);
        };
        /**
         * Update measurements
         */
        ProjectionNode.prototype.updateSnapshot = function () {
            var visible = this.measure();
            this.snapshot = {
                visible: visible,
                layout: this.removeElementScroll(visible),
            };
        };
        ProjectionNode.prototype.updateLayout = function () {
            // TODO: Incorporate into a forwarded
            // scroll offset
            if (this.options.shouldMeasureScroll) {
                this.updateScroll();
            }
            if (!this.isLayoutDirty)
                return;
            this.layout = this.removeElementScroll(this.measure());
            this.layoutCorrected = createBox();
            this.isLayoutDirty = false;
        };
        ProjectionNode.prototype.updateScroll = function () {
            if (!measureScroll)
                return;
            this.scroll = measureScroll(this.instance);
        };
        ProjectionNode.prototype.resetTransform = function () {
            if (resetTransform &&
                this.projectionDelta &&
                (this.isLayoutDirty || this.shouldResetTransform) &&
                !isDeltaZero(this.projectionDelta)) {
                resetTransform(this.instance);
                this.shouldResetTransform = false;
                // TODO: Trigger render to restore transform
            }
        };
        ProjectionNode.prototype.measure = function () {
            if (!measureViewportBox)
                return createBox();
            var box = measureViewportBox(this.instance);
            // Remove window scroll to give page-relative coordinates
            var scroll = this.root.scroll;
            scroll && eachAxis(function (axis) { return translateAxis(box[axis], scroll[axis]); });
            return box;
        };
        ProjectionNode.prototype.removeElementScroll = function (box) {
            var _this = this;
            var boxWithoutScroll = createBox();
            copyBoxInto(boxWithoutScroll, box);
            // TODO: Keep a culmulative scroll offset rather
            // than loop through
            this.path.forEach(function (node) {
                if (node !== _this.root &&
                    node.scroll &&
                    node.options.shouldMeasureScroll) {
                    var scroll_1 = node.scroll;
                    eachAxis(function (axis) {
                        translateAxis(boxWithoutScroll[axis], scroll_1[axis]);
                    });
                }
            });
            return boxWithoutScroll;
        };
        /**
         *
         */
        ProjectionNode.prototype.setTargetDelta = function (delta) {
            this.targetDelta = delta;
            if (!this.projectionDelta)
                this.projectionDelta = createDelta();
            this.root.scheduleUpdateProjection();
        };
        ProjectionNode.prototype.setOptions = function (options) {
            this.options = options;
        };
        /**
         * Frame calculations
         */
        ProjectionNode.prototype.resolveTargetDelta = function () {
            if (!this.targetDelta)
                return;
            if (!this.layout)
                return;
            if (!this.target)
                this.target = createBox();
            copyBoxInto(this.target, this.layout);
            applyBoxDelta(this.target, this.targetDelta);
        };
        ProjectionNode.prototype.calcProjection = function () {
            if (!this.layout || !this.target || !this.projectionDelta)
                return;
            /**
             * Reset the corrected box with the latest values from box, as we're then going
             * to perform mutative operations on it.
             */
            copyBoxInto(this.layoutCorrected, this.layout);
            /**
             * Apply all the parent deltas to this box to produce the corrected box. This
             * is the layout box, as it will appear on screen as a result of the transforms of its parents.
             */
            applyTreeDeltas(this.layoutCorrected, this.treeScale, this.path);
            /**
             * Update the delta between the corrected box and the target box before user-set transforms were applied.
             * This will allow us to calculate the corrected borderRadius and boxShadow to compensate
             * for our layout reprojection, but still allow them to be scaled correctly by the user.
             * It might be that to simplify this we may want to accept that user-set scale is also corrected
             * and we wouldn't have to keep and calc both deltas, OR we could support a user setting
             * to allow people to choose whether these styles are corrected based on just the
             * layout reprojection or the final bounding box.
             */
            calcBoxDelta(this.projectionDelta, this.layoutCorrected, this.target); // origin)
            // TODO Make this event listener
            var onProjectionUpdate = this.options.onProjectionUpdate;
            onProjectionUpdate && onProjectionUpdate();
        };
        ProjectionNode.prototype.getProjectionStyles = function (latest) {
            // TODO: Return lifecycle-persistent object
            var styles = {};
            if (!this.projectionDelta)
                return styles;
            // Resolve crossfading props and viewport boxes
            // TODO: Apply user-set transforms to targetBox
            // TODO: Return persistent mutable object
            styles.transform = buildProjectionTransform(this.projectionDelta, this.treeScale);
            // TODO Move into stand-alone, testable function
            var _a = this.projectionDelta, x = _a.x, y = _a.y;
            styles.transformOrigin = x.origin * 100 + "% " + y.origin * 100 + "% 0";
            /**
             * Apply scale correction
             */
            for (var key in scaleCorrectors) {
                if (latest[key] === undefined) {
                    delete styles[key];
                    continue;
                }
                var _b = scaleCorrectors[key], correct = _b.correct, applyTo = _b.applyTo;
                var corrected = correct(latest[key], this);
                if (applyTo) {
                    var num = applyTo.length;
                    for (var i = 0; i < num; i++) {
                        styles[applyTo[i]] = corrected;
                    }
                }
                else {
                    styles[key] = corrected;
                }
            }
            return styles;
        };
        /**
         * Events
         *
         * TODO Replace this with a key-based lookup
         */
        ProjectionNode.prototype.onLayoutWillUpdate = function (handler) {
            if (!this.layoutWillUpdateListeners) {
                this.layoutWillUpdateListeners = new SubscriptionManager();
            }
            return this.layoutWillUpdateListeners.add(handler);
        };
        ProjectionNode.prototype.onLayoutDidUpdate = function (handler) {
            if (!this.layoutDidUpdateListeners) {
                this.layoutDidUpdateListeners = new SubscriptionManager();
            }
            return this.layoutDidUpdateListeners.add(handler);
        };
        return ProjectionNode;
    }());
}
function updateTreeLayout(node) {
    node.updateLayout();
    node.children.forEach(updateTreeLayout);
}
function notifyLayoutUpdate(node) {
    var layout = node.layout, snapshot = node.snapshot;
    if (layout && snapshot && node.layoutDidUpdateListeners) {
        var layoutDelta = createDelta();
        calcBoxDelta(layoutDelta, layout, snapshot.layout);
        var visualDelta = createDelta();
        calcBoxDelta(visualDelta, layout, snapshot.visible);
        node.layoutDidUpdateListeners.notify({
            layout: layout,
            snapshot: snapshot,
            delta: visualDelta,
            hasLayoutChanged: !isDeltaZero(layoutDelta),
        });
    }
    node.children.forEach(notifyLayoutUpdate);
}
function resetTreeTransform(node) {
    node.resetTransform();
    node.children.forEach(resetTreeTransform);
}
function updateProjectionTree(node) {
    resolveTreeTargetDeltas(node);
    calcTreeProjection(node);
}
function resolveTreeTargetDeltas(node) {
    node.resolveTargetDelta();
    node.children.forEach(resolveTreeTargetDeltas);
}
function calcTreeProjection(node) {
    node.calcProjection();
    node.children.forEach(calcTreeProjection);
}

export { createProjectionNode };
